﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections;
using System.Collections.Generic;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Text;
using HeuristicLab.Common;
using HeuristicLab.Problems.DataAnalysis;

namespace HeuristicLab.Problems.Instances.DataAnalysis {
  public class ClassificationCSVInstanceProvider : ClassificationInstanceProvider {
    public override string Name {
      get { return "CSV File"; }
    }
    public override string Description {
      get {
        return "";
      }
    }
    public override Uri WebLink {
      get { return new Uri("http://dev.heuristiclab.com/trac/hl/core/wiki/UsersFAQ#DataAnalysisImportFileFormat"); }
    }
    public override string ReferencePublication {
      get { return ""; }
    }

    public override IEnumerable<IDataDescriptor> GetDataDescriptors() {
      return new List<IDataDescriptor>();
    }

    public override IClassificationProblemData LoadData(IDataDescriptor descriptor) {
      throw new NotImplementedException();
    }

    public override bool CanImportData {
      get { return true; }
    }
    public override IClassificationProblemData ImportData(string path) {
      TableFileParser csvFileParser = new TableFileParser();

      csvFileParser.Parse(path);

      Dataset dataset = new Dataset(csvFileParser.VariableNames, csvFileParser.Values);
      string targetVar = dataset.DoubleVariables.Last();

      // turn of input variables that are constant in the training partition
      var allowedInputVars = new List<string>();
      var trainingIndizes = Enumerable.Range(0, (csvFileParser.Rows * 2) / 3);
      if (trainingIndizes.Count() >= 2) {
        foreach (var variableName in dataset.DoubleVariables) {
          if (dataset.GetDoubleValues(variableName, trainingIndizes).Range() > 0 &&
            variableName != targetVar)
            allowedInputVars.Add(variableName);
        }
      } else {
        allowedInputVars.AddRange(dataset.DoubleVariables.Where(x => x.Equals(targetVar)));
      }

      ClassificationProblemData classificationData = new ClassificationProblemData(dataset, allowedInputVars, targetVar);

      int trainingPartEnd = trainingIndizes.Last();
      classificationData.TrainingPartition.Start = trainingIndizes.First();
      classificationData.TrainingPartition.End = trainingPartEnd;
      classificationData.TestPartition.Start = trainingPartEnd;
      classificationData.TestPartition.End = csvFileParser.Rows;

      classificationData.Name = Path.GetFileName(path);

      return classificationData;
    }

    protected override IClassificationProblemData ImportData(string path, ClassificationImportType type, TableFileParser csvFileParser) {
      int trainingPartEnd = (csvFileParser.Rows * type.Training) / 100;
      List<IList> values = csvFileParser.Values;
      if (type.Shuffle) {
        values = Shuffle(values);
      }

      Dataset dataset = new Dataset(csvFileParser.VariableNames, values);
      string targetVar = dataset.DoubleVariables.Last();

      // turn of input variables that are constant in the training partition
      var allowedInputVars = new List<string>();
      var trainingIndizes = Enumerable.Range(0, trainingPartEnd);
      foreach (var variableName in dataset.DoubleVariables) {
        if (trainingIndizes.Count() >= 2 && dataset.GetDoubleValues(variableName, trainingIndizes).Range() > 0 &&
          variableName != targetVar)
          allowedInputVars.Add(variableName);
      }

      ClassificationProblemData classificationData = new ClassificationProblemData(dataset, allowedInputVars, targetVar);

      classificationData.TrainingPartition.Start = 0;
      classificationData.TrainingPartition.End = trainingPartEnd;
      classificationData.TestPartition.Start = trainingPartEnd;
      classificationData.TestPartition.End = csvFileParser.Rows;

      classificationData.Name = Path.GetFileName(path);

      return classificationData;
    }

    protected List<IList> Shuffle(List<IList> values, int target, int trainingPercentage, int trainingPartEnd) {
      target = 5;
      IList targetValues = values[target];
      var group = targetValues.Cast<double>().GroupBy(x => x).Select(g => new { Key = g.Key, Count = g.Count() }).ToList();
      Dictionary<double, double> taken = new Dictionary<double, double>();
      foreach (var classCount in group) {
        taken[classCount.Key] = (classCount.Count * trainingPercentage) / 100;
      }

      List<IList> training = GetListOfIListCopy(values);
      List<IList> test = GetListOfIListCopy(values);

      for (int i = 0; i < targetValues.Count; i++) {
        if (taken[(double)targetValues[i]] > 0) {
          AddRow(training, values, i);
          taken[(double)targetValues[i]]--;
        } else {
          AddRow(test, values, i);
        }
      }

      training = Shuffle(training);
      test = Shuffle(test);
      for (int i = 0; i < training.Count; i++) {
        for (int j = 0; j < test[i].Count; j++) {
          training[i].Add(test[i][j]);
        }
      }

      return training;
    }

    private void AddRow(List<IList> destination, List<IList> source, int index) {
      for (int i = 0; i < source.Count; i++) {
        destination[i].Add(source[i][index]);
      }
    }

    private List<IList> GetListOfIListCopy(List<IList> values) {
      List<IList> newList = new List<IList>(values.Count);
      for (int col = 0; col < values.Count; col++) {

        if (values[col] is List<double>)
          newList.Add(new List<double>());
        else if (values[col] is List<DateTime>)
          newList.Add(new List<DateTime>());
        else if (values[col] is List<string>)
          newList.Add(new List<string>());
        else
          throw new InvalidOperationException();
      }
      return newList;
    }

    private List<IList> NormalizeClasses(List<IList> values) {
      int column = GetLastDoubleColumn(values);
      Dictionary<object, int> count = new Dictionary<object, int>();
      foreach (var item in values[column]) {
        if (count.Keys.Contains(item)) {
          count[item]++;
        } else {
          count.Add(item, 1);
        }
      }
      int min = count.Values.Min();
      Dictionary<object, int> taken = new Dictionary<object, int>();
      foreach (var key in count.Keys) {
        taken[key] = 0;
      }
      List<IList> normalizedValues = new List<IList>(values.Count);
      for (int col = 0; col < values.Count; col++) {

        if (values[col] is List<double>)
          normalizedValues.Add(new List<double>());
        else if (values[col] is List<DateTime>)
          normalizedValues.Add(new List<DateTime>());
        else if (values[col] is List<string>)
          normalizedValues.Add(new List<string>());
        else
          throw new InvalidOperationException();
      }
      for (int i = 0; i < values.First().Count; i++) {
        if (taken[values[column][i]] < min) {
          taken[values[column][i]]++;
          for (int col = 0; col < values.Count; col++) {
            normalizedValues[col].Add(values[col][i]);
          }
        }
      }
      return normalizedValues;
    }

    private int GetLastDoubleColumn(List<IList> values) {
      for (int i = values.Count - 1; i >= 0; i--) {
        if (values[i] is List<double>) {
          return i;
        }
      }
      throw new ArgumentException("No possible Target Variable could be found!");
    }

    public override bool CanExportData {
      get { return true; }
    }
    public override void ExportData(IClassificationProblemData instance, string path) {
      var strBuilder = new StringBuilder();
      var colSep = CultureInfo.CurrentCulture.TextInfo.ListSeparator;
      foreach (var variable in instance.Dataset.VariableNames) {
        strBuilder.Append(variable.Replace(colSep, String.Empty) + colSep);
      }
      strBuilder.Remove(strBuilder.Length - colSep.Length, colSep.Length);
      strBuilder.AppendLine();

      var dataset = instance.Dataset;

      for (int i = 0; i < dataset.Rows; i++) {
        for (int j = 0; j < dataset.Columns; j++) {
          if (j > 0) strBuilder.Append(colSep);
          strBuilder.Append(dataset.GetValue(i, j));
        }
        strBuilder.AppendLine();
      }

      using (var writer = new StreamWriter(path)) {
        writer.Write(strBuilder);
      }
    }
  }
}
